# Author:: Gerhard Spitzlsperger
# EMAIL and WWW:: ruby@gs68.de, www.gs68.de
# Copyright:: Copyright (c) 2003 Gerhard Spitzlsperger. All rights reserved.
# License:: LGPL see file LICENSE.
# Version 0.1

require 'fox'
require 'fox/colors'

require 'GS_PlotItems'

include Fox

# A little drawing canvas to use in our tests for float coordinates
class GS_PlotCanvas < FXCanvas
  attr_reader :items, :x_min, :x_max, :y_min, :y_max 
  attr_reader :left_margin, :right_margin, :bottom_margin, :top_margin
  attr_reader :x_scale, :y_scale;

  def initialize(p, tgt, sel, opts, area)
    # Invoke base class initialize function first
    super(p, tgt, sel, opts)  

    # plotting area
    @x_min = area[0]
    @y_min = area[1]
    @x_max = area[2]
    @y_max = area[3]
     
    # margins on the canvas side 
    @left_margin = 50
    @right_margin = 50
    @top_margin = 50
    @bottom_margin = 50
    
    # usually x, and y are assumed to be linear axis
    @x_scale = "lin"
    @y_scale = "lin"
     
    # hash to hold elements which are plotted
    @items = Hash.new("not existing")
  end
  
  # append new items to plot (name must be unique otherwise it
  # will be overridden
  def append_item(name, item)
     #print "append #{name} \n"
     @items.update({name => item})
  end
       
  # plots everything on the canvas
  def draw
    dc = GS_PlotContext.new(self)
    dc.plot
  end  
  
  def getArea()
    return [@x_min, @x_max, @y_min, @y_max]
  end  
  
  def pr()
    p "now printing items:\n"
  
    @items.each_key{ |item| 
      p item
  }
  end  
  
  def setArea(area)
    # plotting area
    
    begin
      unless area[0] < area[1] && area[2] < area[3]
        Kernel::raise DrawAreaException, "wrong setting of draw area"
      end	
    end
 
    @x_min = area[0]
    @x_max = area[1]
    @y_min = area[2]
    @y_max = area[3]
  
    # redraw
    #draw ## this crashes some applications
 
    # return new area
    getArea
  end    
  
  def setMargin(l, r, t, b)
    # plotting area
    
    begin
      unless l >= 0 && r >= 0 && t >= 0 && b >= 0 
        Kernel::raise DrawAreaException, "wrong setting of draw area"
      end	
    end
    
    # margins on the canvas side 
    @left_margin = l
    @right_margin = r
    @top_margin = t
    @bottom_margin = b    
  end 
  
  def setScale(x, y)
    @x_scale = x
    @y_scale = y
  end       
    
end

class GS_PlotContext < FXDCWindow
  def initialize(pcanvas)
    # Invoke base class initialize function first
    super(pcanvas)

    # keep the connection to the plot canvas
    @pc = pcanvas
  end
  
  def plot()   
    # prepare the drawing area (erase it)
    self.foreground = FXColor::White
    fillRectangle(0, 0, @pc.getWidth(), @pc.getHeight())
    
    # Set the foreground color for drawing
    self.foreground = FXColor::Black

    # restrict drawing area
    setClipRectangle(@pc.left_margin, @pc.top_margin,
                     @pc.getWidth() - @pc.right_margin - @pc.left_margin,
                     @pc.getHeight() - 	@pc.top_margin - @pc.bottom_margin)	     
    
    # for ploting the values will be sorted according depth therefore first an array of 
    # items is setup which is sorted in the second step
    a = @pc.items.values
    a.sort! {|i,j| i.depth <=> j.depth }
    
    a.each { |item| 
      # set the plot style
      
      item.baseStyle ? item.baseStyle.adjustDC(self) : GS_Style.new.adjustDC(self) 

      if item.isVisible == true
        # prepare for exceptions
        begin 
          # check which item to plot
          if item.type == "GS_Line"
            plotLine(item)        
      	  elsif item.type == "GS_Function"
            plotFunction(item)
          elsif item.type =~ /Axis/
      	    plotAxis(item)
      	  elsif item.type == "GS_Rectangle"
            plotRectangle(item)	   
      	  elsif item.type == "GS_Text"
            plotText(item)	  
          else
            Kernel::raise UnexpectedPlotItemException, 
    	      "unexpected item to plot"
          end
        end # exception handling
      end	
    }
    
    # reset drawing area
    clearClipRectangle    
  end
  
  # transforms real coordinates to screen coordinates
  protected
  def transform( x, y )
     x = x.to_f
     y = y.to_f
     
     if x.nan? || y.nan? || x.infinite? != nil || y.infinite? != nil 
      Kernel::raise CannotConvertException, 
        "point to convert is nan, Point(#{x}, #{y})" 
    end
  
    rw = 1.0 * @pc.x_max - @pc.x_min
    rh = 1.0 * @pc.y_max - @pc.y_min
    
    # account for edges
    pixel_width = @pc.getWidth() - @pc.left_margin - @pc.right_margin
    pixel_height = @pc.getHeight() - @pc.bottom_margin - @pc.top_margin
    
    # calculate screen coordinates
    xp =  @pc.left_margin + pixel_width * (x -  @pc.x_min) / rw
    yp =  @pc.top_margin + pixel_height - pixel_height * (y - @pc.y_min) / rh

    if xp.nan? || yp.nan? || xp.infinite? != nil || yp.infinite? != nil 
      Kernel::raise CannotConvertException, 
    	"abnormal conversion"
    end

    return [xp, yp]
  end

  # scales a real x distnce into pixels
  def scaleX ( x )
     rw = 1.0 * @pc.x_max - @pc.x_min
     pixel_width = @pc.getWidth() - @pc.left_margin - @pc.right_margin
     
     return x * pixel_width / rw
  end

  # scales a real y distnce into pixels
  def scaleY ( y )
     rh = 1.0 * @pc.y_max - @pc.y_min
     pixel_height = @pc.getHeight() - @pc.bottom_margin - @pc.top_margin
     
     return y * pixel_height / rh
  end

  # plot an axis
  def plotAxis(item)  
    if item.type =~ /Log/
      plotLogAxis(item)
    else
      plotLinAxis(item)
    end      
  end

  # plots a function item
  protected 
  def plotFunction(item)

    # individual points must be copied to ovoid overriding of initial data
    # in case of log plots negatve y values are replaced by minimum positive numbers!!!!
    if @pc.x_scale == "log" && @pc.y_scale == "log"
         temp = item.data.map { |o| p=o.dup;  p.setX(Math.log10(p.x)); 
	                            t = p.y <= 0.0 ? 1.0e-300 : p.y; 
				    p.setY(Math.log10(t)); p }
    elsif @pc.x_scale == "log" && @pc.y_scale == "lin"
         temp = item.data.map { |o| p=o.dup;  p.setX(Math.log10(p.x)); p }	 
    elsif @pc.x_scale == "lin" && @pc.y_scale == "log"
         temp = item.data.map { |o| p=o.dup;  
	                            t = p.y <= 0.0 ? 1.0e-300 : p.y; 
				    p.setY(Math.log10(t)); p }
    else
         temp = item.data    
    end
    
    # set up array of FXPoints to draw the lines
    points = []
    
    temp.each { |p| 
      a = transform(p.x, p.y)
      points << FXPoint.new(a[0], a[1])
    }
    
    drawLines(points)
  end

  #plot an linear axis
  def plotLinAxis(item)
    # check the size for autoscale axis
    if item.type == "GS_XAxis_Autoscale"
      item.setX(@pc.x_min, @pc.x_max, nil)
    end
    
    if item.type == "GS_YAxis_Autoscale"
      item.setY(@pc.y_min, @pc.y_max, nil)
    end
    
    # draw the main line
    a = transform(item.x1, item.y1)
    b = transform(item.x2, item.y2) 
     
    drawLine(a[0], a[1], b[0], b[1])
    
    #!!! the following is bad style but fast to implement
    return if( !item.ticks_and_numbers )

    if item.type =~ /GS_XAxis/
      item.major_tick_pos.each { |pos|
        a = transform(pos, item.y1)
        drawLine(a[0], a[1]-item.tick_size, a[0], a[1] + item.tick_size)
	
	# draw axislabel
	pos_s = item.calcLabel(pos, item.x1, item.x2)
	unless (pos.abs < item.major_tick * 0.2 && item.omit_0_label )
	  drawText(a[0] + item.num_tick_offset, 
	           a[1] + item.tick_size + item.x_num_dist, pos_s ) 
	end
      }
    else
      item.major_tick_pos.each { |pos|
        a = transform(item.x1, pos)
        drawLine(a[0] - item.tick_size, a[1], a[0] + item.tick_size, a[1])
	
	pos_s = item.calcLabel(pos, item.y1, item.y2)
	unless (pos.abs < item.major_tick * 0.2 && item.omit_0_label )
	  drawText(a[0] + item.tick_size + item.y_num_dist, 
	           a[1] - item.num_tick_offset, pos_s)
	end	   
      }    
    end
  end
   
  # plots a line item
  protected 
  def plotLine(item)
    a = transform(item.x1, item.y1)
    b = transform(item.x2, item.y2)
    drawLine(a[0], a[1], b[0], b[1])
  end
  
  #plot an axis
  protected
  def plotLogAxis(item)
    # check the size for autoscale axis
    if item.type == "GS_LogXAxis_Autoscale"
      item.setX(@pc.x_min, @pc.x_max, nil)
    end
    
    if item.type == "GS_LogYAxis_Autoscale"
      item.setY(@pc.y_min, @pc.y_max, nil)
    end
     
    a = transform(item.x1, item.y1)
    b = transform(item.x2, item.y2) 
     
    drawLine(a[0], a[1], b[0], b[1])
    
    #!!! the following is bad style but fast to implement
    return if( !item.ticks_and_numbers )
    
    if item.type =~ /GS_LogYAxis/
      item.major_tick_pos.each { |pos|
        a = transform(item.x1, pos)
	# everything is now in window coordinates
        drawLine(a[0] - item.tick_size, a[1], a[0] + item.tick_size, a[1])
	
	pos_s = item.calcLabel(pos, item.y1, item.y2)
	drawText(a[0] + item.tick_size + item.y_num_dist, 
	         a[1] - item.num_tick_offset, pos_s)
		 
	item.minor_tick_pos.each{ |mpos|
	  a = transform(item.x1, pos + Math.log10(mpos))
	  if( a[1] >= @pc.top_margin + 5 ) # 
            drawLine(a[0] - item.minor_tick_size, a[1], a[0] + item.minor_tick_size, a[1])
	  end  
	}
      } 
    elsif item.type =~ /GS_LogXAxis/
      item.major_tick_pos.each { |pos|
        a = transform(pos, item.y1)
	# everything is now in window coordinates
        drawLine(a[0], a[1] - item.tick_size, a[0], a[1] + item.tick_size)
	
	pos_s = item.calcLabel(pos, item.x1, item.x2)
	drawText(a[0] + item.num_tick_offset,
	         a[1] + item.tick_size + item.x_num_dist, pos_s)
		 
	item.minor_tick_pos.each{ |mpos|
	  a = transform(pos + Math.log10(mpos), item.y1 )
	  if( a[0] >= @pc.left_margin + 5 )
            drawLine(a[0], a[1] - item.minor_tick_size, a[0], a[1] + item.minor_tick_size)
	  end  
	}
      } 
    end
  end
   
  # plots a rectangle
  protected 
  def plotRectangle(item)
    a = transform(item.x, item.y)
    w = scaleX(item.w)
    h = scaleY(item.h)
  
    self.foreground = FXColor::White
    fillRectangle(a[0], a[1], w, h)
    
    # Set the foreground color for drawing
    self.foreground = FXColor::Black

    drawRectangle(a[0], a[1], w, h)
  end
     
   
# plots a text item
  protected 
  def plotText(item)
    a = transform(item.x, item.y)

    drawText(a[0], a[1], item.text)
  end
end

# define exceptions
class DrawAreaException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end

# define exceptions
class UnexpectedPlotItemException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end

# define exceptions
class CannotConvertException < RuntimeError
  attr :message
  def initialize(m)
    @message = m
    
    print "!! " + @message + " !!\n"
  end
end