# Author:: Gerhard Spitzlsperger
# EMAIL and WWW:: ruby@gs68.de, www.gs68.de
# Copyright:: Copyright (c) 2003 Gerhard Spitzlsperger. All rights reserved.
# License:: LGPL see file LICENSE.
# Version 0.1

require 'GS_AreaDialog'
require 'GS_PlotCanvas'
require 'GS_FramePlotCanvas'

require 'fox'
require 'fox/colors'
require 'GS_CanvasAsDialog'
require 'GS_FlashCell'
require 'GS_Physics'
require 'GS_Semiconductor'

include Fox
include GS_Physics
include GS_Semiconductor


class NumericInputField
  attr_reader   :name
  attr_reader   :inputRange # should be an array with two elements
  attr_reader   :lname
  attr_reader   :unit
  
  def initialize(n, r, l, u)
    @name = n
    @inputRange = r
    @lname = l
    @unit = u
  end
  
  def to_s
    return @name + " " + @inputRange[0] + ".." + @inputRange[1] +
        @lname + " " + @unit  
  end  
end

class CanvasWindow < FXMainWindow  

  def initialize(app)
    # Call base class initializer first
    super(app, "Very Simple Flash Cell Simulator", nil, nil, DECOR_ALL,
      0, 0, 950, 600)
    
    # init the menues and the toolbar
    initMenues 
    initToolbar
    
    # contains the canvas
    @canvasFrame = FXVerticalFrame.new(self,
      FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,
      0, 0, 0, 0, 0, 0, 0, 0)  

    # contains the input canvas
    @inputFrame = FXHorizontalFrame.new(@canvasFrame,
      FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,
      0, 0, 0, 0, 0, 0, 0, 0)

    # Drawing canvas
    @canvas = GS_PlotCanvas.new(@inputFrame, nil, 0, (FRAME_SUNKEN|FRAME_THICK|
      LAYOUT_FILL_Y|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT),[-1.8, -1.0, 1.0, 1.0])
      
    @canvas.setMargin(25, 25, 25, 25 )  
    @canvas.resize(800, 500)
 
    @canvas.connect(SEL_PAINT) { |sender, sel, event|
      @canvas.draw
    }
    
    # init the input area (must be called before initMemorySketch
    initInputArea

    # init the memory sketch
    initMemorySketch

    # the FETMOS
    @fmos = FETMOS.new    
  end
  
  def initInputArea
    # now create the input area on the right side and arange nicely
    # do never change order of inputs its assuemd to be fixed always
 
    # setup the nodes of the device with the input voltage range
    # @nodes are also inputs, but they are handled slightly different
    # sometimes
    @nodes = %w(cg sub drain field source)    
    @inputs = Array.new
    
    @inputs.push(["cg",
        NumericInputField.new("cg", [-20.0, 20.0], "Control Gate", "V")]
	)	
    @inputs.push(["sub",
        NumericInputField.new("sub", [-20.0, 20.0], "Substrate", "V")]
	) 
    @inputs.push(["drain",
        NumericInputField.new("drain", [-20.0, 20.0], "Drain", "V")]
	)
    @inputs.push(["field",
        NumericInputField.new("field", [-20.0, 20.0], "Field", "V")]
	) 
    @inputs.push(["source", 
        NumericInputField.new("source", [-20.0, 20.0], "Source", "V")]
	)

    @inputs.push(["fg_charge", 
        NumericInputField.new("fg_charge", [0, 100.0], "Floating Gate Charge", "fC")]
	)	
    @inputs.push(["channel_width",
        NumericInputField.new("channel_width", [0.0, 2.0], "Channel Width", "um")]
	) 
    @inputs.push(["channel_length",
        NumericInputField.new("channel_length", [0.0, 2.0], "Channel Length", "um")]
	)
    @inputs.push(["cell_length", 
        NumericInputField.new("cell_length", [0, 5.0], "Cell Length", "um")]
	)	
    @inputs.push(["drain_over",
        NumericInputField.new("drain_over", [0.0, 100.0], "Drain Overlap", "%")]
	) 
    @inputs.push(["on_c",
        NumericInputField.new("on_c", [0.0, 50.0], "ON Capacity", "fF/um^2")]
	)
    @inputs.push(["fto_t",
        NumericInputField.new("fto_t", [0.0, 50.0], "FTO thickness", "nm")]
	)
    
    # we use a FXDataTarget to keep sliders and textfield aligned
    @dataTargets = Hash.new  
    @textFields = Hash.new   
    @sliders = Hash.new      

    vFrame = FXVerticalFrame.new(@inputFrame,
      FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,
      0, 0, 0, 0, 0, 0, 0, 0) 
      
    matrix = FXMatrix.new(vFrame, 4, MATRIX_BY_COLUMNS)

    @inputs.each { |item| # item is a array with wo elements:
                          # name and numericInputField  
      @dataTargets[item[0]] = FXDataTarget.new(
        (item[1].inputRange[0] + item[1].inputRange[1]) / 2.0
      )			  
			  
      FXLabel.new(matrix, item[1].lname, nil,
        LAYOUT_CENTER_Y|LAYOUT_CENTER_X|JUSTIFY_RIGHT|LAYOUT_FILL_ROW)
      FXLabel.new(matrix, item[1].unit, nil,
        LAYOUT_CENTER_Y|LAYOUT_CENTER_X|JUSTIFY_RIGHT|LAYOUT_FILL_ROW)  
	
      @textFields[item[0]] = FXTextField.new(matrix, 10, @dataTargets[item[0]], 
        FXDataTarget::ID_VALUE,
        (TEXTFIELD_REAL|LAYOUT_CENTER_Y|LAYOUT_CENTER_X|
        FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_ROW))
      
      @sliders[item[0]] = FXSlider.new(matrix, @dataTargets[item[0]], 
        FXDataTarget::ID_VALUE,
        LAYOUT_CENTER_Y|LAYOUT_FILL_X|LAYOUT_FILL_ROW|LAYOUT_FIX_WIDTH, 0, 0, 100)

      @sliders[item[0]].range = item[1].inputRange[0]..item[1].inputRange[1]
    } 
    
    # add the buttons 
    FXHorizontalSeparator.new(vFrame,
      LAYOUT_SIDE_TOP|SEPARATOR_GROOVE|LAYOUT_FILL_X)  

    calcButton = FXButton.new(vFrame,
      "calc ...", nil, nil, 0,
      FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X|LAYOUT_CENTER_Y)
    calcButton.connect(SEL_COMMAND) { |sender, sel, event|
         calcPhiFG
      }
      
    simuButton = FXButton.new(vFrame,
      "simulate ...", nil, nil, 0,
      FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X|LAYOUT_CENTER_Y)
    simuButton.connect(SEL_COMMAND) { |sender, sel, event|
         simulate
      }
  end

  def initMemorySketch
    # create the drawing canvas with content
    size = 0.6
    step = 0.75
    setupCapacitor(@nodes[0], 0.0, 0.5, size )
    @canvas.append_item( @nodes[0] + "V" , 
      GS_Text.new( 0.0, 0.4+  size*0.8, "init"))
  
    setupCapacitor(@nodes[1], -1.0*step, -0.5, size )
    @canvas.append_item( @nodes[1] + "V" , 
      GS_Text.new( -1.0*step, -0.5 - size*0.8, "init"))
      
    setupCapacitor(@nodes[2], 0.0, -0.5, size )
    @canvas.append_item( @nodes[2] + "V" , 
      GS_Text.new( 0.0, -0.5 - size*0.8, "init"))    

    setupCapacitor(@nodes[3], step, -0.5, size )
    @canvas.append_item( @nodes[3] + "V" , 
      GS_Text.new( step, -0.5 - size*0.8, "init"))
      
    setupCapacitor(@nodes[4], -2.0*step, -0.5, size )
    @canvas.append_item( @nodes[4] + "V" , 
      GS_Text.new( -2.0*step, -0.5 - size*0.8, "init"))

    gl = GS_Line.new( -2.0*step, 0.0, step, 0.0 )    
    @canvas.append_item( "horizontal", gl )
    
    gl = GS_Line.new( -1.0*step, 0.0, -1.0*step, -0.3 )    
    @canvas.append_item( "vertical_left", gl )
    
    gl = GS_Line.new( -2.0*step, 0.0, -2.0*step, -0.3 )    
    @canvas.append_item( "vertical_veryleft", gl )   

    gl = GS_Line.new( step, 0.0, step, -0.3 )    
    @canvas.append_item( "vertical_right", gl )
    
    gl = GS_Line.new( 0.0, 0.3, 0.0, -0.3 )    
    @canvas.append_item( "vertical", gl )
    
    # the rectangle for the floating gate
    gr = GS_Rectangle.new( -0.5, 0.16, 1.0, 0.32)
    gr.depth = 10
    @canvas.append_item( "FG_Rect" , gr ) 
    
    # the potential indicator
    gt = GS_Text.new( -0.45, 0.05, "FG: phi")
    gt.depth = 20    
    @canvas.append_item( "FG_V" , gt )
    
    # the charge indicator
    gt = GS_Text.new( -0.45, -0.05, "FG: charge")
    gt.depth = 20    
    @canvas.append_item( "FG_C" , gt )
  end
  
  def initMenues
    # Make menu bar
    menubar = FXMenubar.new(self, LAYOUT_FILL_X)
    
    # the file menu
    filemenu = FXMenuPane.new(self)  
    FXMenuTitle.new(menubar, "&File", nil, filemenu)
    FXMenuCommand.new(filemenu, "&Quit\tCtl-Q", nil,
      getApp(), FXApp::ID_QUIT)
     
    # the help menu
    helpmenu = FXMenuPane.new(self)
    FXMenuTitle.new(menubar, "&Help", nil, helpmenu, LAYOUT_RIGHT)
    FXMenuCommand.new(helpmenu, "&About ...").connect(SEL_COMMAND) {
      FXMessageBox.information(self, MBOX_OK, 
        "About vsfcs.rb (Very Simple Flash Cell Simulator)",
        "A simple program to calculate floating gate potentials \n" +
        "for Flash Cells based on the capacitor model (www.gs68.de)\n\n" +
	"Copyright 2003 gerhard.spitzlsperger@gs68.de")
    }
  end

  def initToolbar
    # load some icons  
    killicon = getIcon("kill.png")
  
    # Place a toolbar above the canvas
    toolbar = FXToolbar.new(self, (LAYOUT_SIDE_TOP|
      PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT|FRAME_RAISED|LAYOUT_FILL_X))
 
    FXButton.new(toolbar, "Exit\tExit", killicon, app, FXApp::ID_QUIT,
      ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED)     
  end

  def calcPhiFG
      fgt = 0.2 # gate height in um (assumed fixed)
  
      cl = @textFields['channel_length'].getText.to_f # in um
      cw = @textFields['channel_width'].getText.to_f # in um
      cel = @textFields['cell_length'].getText.to_f # in um
      dover = @textFields['drain_over'].getText.to_f  # in %
      c_a_cg = @textFields['on_c'].getText.to_f # in fF/um^2
      t_fto = @textFields['fto_t'].getText.to_f #in nm
      q_fg = @textFields['fg_charge'].getText.to_f # in fC

      # put the input potentials to the sketch and the @fmos
      @nodes.each { |item|
      	v = @textFields[item].getText
	@canvas.items[item + "V"].text = v + " V"
	@fmos.setPhi(item, v.to_f)
      }
      
      # setting the values -> calculates the new potential and in the flash cell
      @fmos.setFTO( t_fto, cl * cw, dover )
      @fmos.setCG( c_a_cg, cl * cel )
      @fmos.set_q_FG( q_fg )
      
      # displays the claculated set the format string depending on the number to be displayed
      fs = (@fmos.phi_fg.abs >= 10.0 ? "FG: %.4g V" : "FG: %.3g V")
      @canvas.items["FG_V"].text = sprintf(fs, @fmos.phi_fg)
      
      fs = (@fmos.q_fg.abs >= 10.0 ? "FG: %.4g fC or %d q" : 
                                     "FG: %.3g fC or %d q")
      @canvas.items["FG_C"].text = sprintf(fs, @fmos.q_fg, @fmos.q_fg * 1e-15 / Q)     

      # set the new values
      @nodes.each { |item|
	c = @fmos.getVoltage(item)
	
	# set the format string depending on the number to be displayed
	fs = (c.abs >= 10.0 ? "%.4g V" : "%.3g V")
	@canvas.items[item + "voltage"].text = sprintf(fs,c)

	c = @fmos.c[item]
	# set the format string depending on the number to be displayed
	fs = (c.abs >= 10.0 ? "%.4g fF" : "%.3g fF")
	@canvas.items[item + "capacity"].text = sprintf(fs,c)
      }
    
      @canvas.draw
  end
  
  # do the tunneling simulation
  def simulate
    r = @fmos.simulate  
    
    # build a canvas dialog box
    rc = GS_CanvasAsDialog.new(self)
    
    gf = GS_Function.new(0.0, 0.0001, 43, r)
    rc.canvas.append_item( "test2", gf )
    
    a = [0.0, 0.0001, -200, 100]
    rc.canvas.setArea(a)

    rc.execute
  end
  
  # support function to draw the capacitors on the memory sketch
  def setupCapacitor(name, x, y, size)
    # move x, y to left corner
    x = x - size/2.0
    y = y - size/2.0
    
    gl = GS_Line.new( x, y + 0.45*size, x + size, y + 0.45*size )    
    @canvas.append_item( name + "upper plate", gl )
    
    gl = GS_Line.new( x, y + 0.55*size, x + size, y + 0.55*size )    
    @canvas.append_item( name + "lower plate", gl )
    
    gl = GS_Line.new( x + size/2.0, y + 0.45*size, x + size/2.0, y)    
    @canvas.append_item( name + "upper connect", gl )
    
    gl = GS_Line.new( x + size/2.0, y + 0.55*size, x + size/2.0, y + size )    
    @canvas.append_item( name + "lower connect", gl )
    
    gt = GS_Text.new( x, y + 0.9*size, name)
    @canvas.append_item( name , gt )
    
    gt = GS_Text.new( x, y, "undef F")
    @canvas.append_item( name + "capacity" , gt )
    
    gt = GS_Text.new( x, y + 0.2*size, "undef V")
    @canvas.append_item( name + "voltage", gt )
  end

  #--
  #-- GUI related support functions
  #--
  def getIcon(filename)
    FXPNGIcon.new(getApp(), File.open(filename, "rb").read)
  end
  
  # Create and show the main window
  def create
    super                  # Create the windows
    show(PLACEMENT_SCREEN) # Make the main window appear
  end
end

#--
#-- the main routine
#--

def run
  # Construct the application object
  application = FXApp.new('vsfcs', 'vsfcs')
  
  # Initialize it
  application.init(ARGV)
  
  # Construct the main window
  s = CanvasWindow.new(application)
  
  # Create the application
  application.create
  
  # Run the application
  application.run
end

# run the application
run


